#include "settings.lua"

function init()
  
  debug = false
  
  -- settings check
  if not HasKey(SETTINGS_BASE) then
    SetDefaults()
  end
  checkSettingsVersion()
  checkKeybinds()
  checkGroup()

  showPopup = GetBool(SETTINGS_BASE .. ".popup")
  plankGroup  = GetInt(SETTINGS_BASE .. ".group")

  keyPressedLast = nil
  keyPressedLastTime = 0
  keyPressedLastTimeWait = 0.35
  keyPressedLastTimeMax = 0.07
  keyPressedLastWaiting = true

  plankStart = nil
  plankStartShape = nil
  plankStartNormal = nil
  plankStartDistance = nil
  plankTooClose = false
  plankEnd = nil
  plankEndNormal = Vec(0,1,0)
  plankConnects = false
  plankWidth = GetInt(SETTINGS_BASE .. ".width")
  plankWidthMin = 1
  plankWidthMax = 50
  plankHeight = GetInt(SETTINGS_BASE .. ".height")
  plankHeightMin = 1
  plankHeightMax = 50
  soundError = LoadSound('error.ogg')
  errorPress = false
  couldPlace = false
  placeDistance = 3
  frontDistance = 1
  plankStrength = 100
  minimumLength = 0.1
  toolOffset = nil
  toolOffsetHidden = Vec(0.25,-1.6,-0.3)
  toolOffsetNormal = Vec(0.25,-0.65,-0.3)
  toolSpeed = 0.4
  soundsAttach = {}
  for i=1,6 do
    table.insert(soundsAttach,LoadSound("tools/plank-attach"..(i-1)..".ogg"))
  end
  soundsDetach = {}
  for i=1,9 do
    table.insert(soundsDetach,LoadSound("tools/plank-detach"..(i-1)..".ogg"))
  end
  
  plankStyles = {
    {
      name="Wood",
      mat="wood",
      path="MOD/vox/plank_wood.vox",
      brush="MOD/vox/plank_wood_brush.vox",
      color="0.75 0.7 0.65",
      tags="RHA=80",
      strength="1"
    },
    {
      name="Metal",
      mat="metal",
      path="MOD/vox/plank_metal.vox",
      brush="MOD/vox/plank_metal_brush.vox",
      color="0.7 0.7 0.7",
      tags="RHA=200",
      strength="10"
    },
    {
      name="Glass",
      mat="glass",
      path="MOD/vox/plank_glass.vox",
      brush="MOD/vox/plank_glass_brush.vox",
      color="0.7 0.7 0.7 0.5",
      tags="RHA=10",
      strength="1"
    },
    {
      name="Plastic",
      mat="plastic",
      path="MOD/vox/plank_plastic.vox",
      brush="MOD/vox/plank_plastic_brush.vox",
      color="1 1 1",
      tags="RHA=30",
      strength="1"
    },
    {
      name="Concrete",
      mat="concrete",
      path="MOD/vox/plank_concrete.vox",
      brush="MOD/vox/plank_concrete_brush.vox",
      color="1 1 1",
      tags="RHA=150",
      strength="1"
    }
  }
  plankIndex = GetInt(SETTINGS_BASE .. ".style")
  switchBack = false
  
  if plankGroup == 0 then
    plankGroup = 2
    RegisterTool("infiniteplank", "Plank+", plankStyles[plankIndex].path, plankGroup)
    SetBool("game.tool.plank.enabled", false)
  else
    RegisterTool("infiniteplank", "Infinite Plank", plankStyles[plankIndex].path, plankGroup)
  end
  SetBool("game.tool.infiniteplank.enabled", true)

end

function Style(idx)
  local style = plankStyles[idx]
  plankIndex = idx
  SetInt(SETTINGS_BASE .. ".style", plankIndex)
  
  if GetInt(SETTINGS_BASE .. ".group") == 0 then
    RegisterTool("infiniteplank", "Plank+", style.path, 2)
  else
    RegisterTool("infiniteplank", "Infinite Plank", style.path, plankGroup)
  end
  
  SetBool("game.tool.plank.enabled", true)
  SetString("game.player.tool","plank")
  switchBack = true
  SetString("hud.notification",style.name)
end

function PlayRandomSound(type)
  if type == "start" or type == "could" then
    PlaySound(soundsAttach[math.random(1,#soundsAttach)])
  end
  if type == "end" then
    PlaySound(soundsDetach[math.random(1,#soundsDetach)])
  end
  if type == "error" then
    PlaySound(soundError)
  end
end

function tick(dt)
  if debug then -- debug
    SetString("game.player.tool","infiniteplank")
  end
  
  if switchBack then
    SetString("game.player.tool","infiniteplank")
    if GetInt(SETTINGS_BASE .. ".group") == 0 and GetBool("game.tool.plank.enabled") then -- make sure plank is disabled
      SetBool("game.tool.plank.enabled", false)
    end
    switchBack = false
  end

  if GetInt(SETTINGS_BASE .. ".group") == 0 and GetBool("game.tool.plank.enabled") then -- make sure plank is disabled
    SetBool("game.tool.plank.enabled", false)
  end
  
	if GetString("game.player.tool") == "infiniteplank" then 
    
    -- check if holding down key
    if keyPressedLast then
      keyPressedLastTime = keyPressedLastTime + dt
      if keyPressedLastTime > keyPressedLastTimeWait then
        keyPressedLastWaiting = false
      end
      local max = keyPressedLastTimeMax
      if keyPressedLastWaiting then
        max = keyPressedLastTimeWait
      end
      -- print(keyPressedLastTime)
      if keyPressedLastTime > max then
        keyPressedLastTime = 0
        if KeyDownCheck(keyPressedLast) then
          KeyPressed(keyPressedLast)
        else
          keyPressedLast = nil
        end
      end
    end
    
    if plankStart then -- created transformed versions of plankstart and plankstartnormal if plankstart so that we don't need to create them mutliple times per frame
  	   UpdateTransformedStart()
    end

    if showPopup and string.lower(InputLastPressedKey()) == "y" then
      SetInt(SETTINGS_BASE .. ".group", 0)
      SetBool(SETTINGS_BASE .. ".popup", false)
      showPopup = false
      RegisterTool("infiniteplank", "Plank+", plankStyles[plankIndex].path, 2)
    	SetBool("game.tool.infiniteplank.enabled", true)
      SetBool("game.tool.plank.enabled", false)
      SetString("game.player.tool","infiniteplank")
    end

    if showPopup and string.lower(InputLastPressedKey()) == "n" then
      SetBool(SETTINGS_BASE .. ".popup", false)
      showPopup = false
    end


    if KeyPressCheck("switch") then
      plankIndex = plankIndex + 1
      if plankIndex > #plankStyles then
        plankIndex = 1
      end
      Style(plankIndex)
    end
    
    if KeyPressCheck("width_add") or KeyPressCheck("width_sub") or KeyPressCheck("height_add") or KeyPressCheck("height_sub") then
      KeyPressed()
    end
    
    if GetBool("game.player.canusetool") then -- main body of tool functionality
      local b = GetToolBody()
      if b ~= 0 then
        if not toolOffset then
          toolOffset = toolOffsetHidden
        end
        if plankStart then
            toolOffset = VecLerp(toolOffset,toolOffsetHidden,toolSpeed)
        else
          toolOffset = VecLerp(toolOffset,toolOffsetNormal,toolSpeed)
        end
        SetToolTransform( Transform(toolOffset,QuatEuler(0,-18,-2)), 1 )
      end
      if InputDown("usetool") and not plankStart and not errorPress then -- start plank placing when lmb is pressed
        local t = GetCameraTransform()
        local dir = TransformToParentVec(t, Vec(0, 0, -1))
        local hit, dist, normal, shape = QueryRaycast(t.pos, dir, placeDistance)
        if hit then
            local hitpoint = VecAdd(t.pos, VecScale(dir, dist))
            local shapetf = GetShapeWorldTransform(shape)
            plankStartShape = shape
            plankStart = TransformToLocalPoint(shapetf, hitpoint)
            plankStartNormal = TransformToLocalVec(shapetf, normal)
            plankStartDistance = dist + 0.3
            couldPlace = true
            PlayRandomSound("start")
            UpdateTransformedStart()
        else
            errorPress = true
            PlayRandomSound("error")
            PlankReset()
        end
  		end
      if InputReleased("usetool") then -- if lmb is released then place plank
        errorPress = false
        if plankStart then
            if not plankTooClose then
              SpawnPlank(plankStartTransformed,plankEnd,plankEndNormal)
              PlayRandomSound("end")
            else
              PlayRandomSound("error")
            end
            PlankReset()
        else
          PlankReset()
        end
      end
    else
      SetToolTransform( Transform(toolOffset,QuatEuler(0,-18,-2)), 1 )
    end
    if InputDown("usetool") and plankStart then
      local t = GetCameraTransform()
      local dir = TransformToParentVec(t, Vec(0, 0, -1))
      local hitpoint = VecAdd(t.pos,VecScale(dir,plankStartDistance))

      local hit, dist, normal, shape = QueryRaycast(t.pos, dir, plankStartDistance + 0.1)
      if hit then
        hitpoint = VecAdd(t.pos, VecScale(dir, dist))
        
        if not couldPlace then
          PlayRandomSound("could")
        end
        couldPlace = true
        plankEndNormal = GetCorrectPlankNormal(normal)
        plankConnects = true
      else
        plankEndNormal = Vec(0,0,0)
        if couldPlace then
          PlayRandomSound("end")
        end
        couldPlace = false
        plankConnects = false
      end

      local mag = VecLength(VecSub(plankStartTransformed, hitpoint)) 
      plankEnd = VecAdd(plankStartTransformed, VecScale(VecScale(VecSub(hitpoint, plankStartTransformed),1/mag),math.max(0.1,math.round(mag*10)/10)))

      -- check if point are very close
      if mag < 0.001 then
        plankTooClose = true
      else
        plankTooClose = false
        DrawPlankPreview()
      end
    end
  else
    PlankReset()
	end
end


function PlankReset()
  plankStart = nil
  plankStartShape = nil
  plankStartNormal = nil
  plankStartDistance = nil
  plankTooClose = false
  plankEnd = nil
  plankEndNormal = Vec(0,1,0)
  couldPlace = false
  plankConnects = false
end


function DrawPlankPreview()
  local normal = GetCorrectPlankNormal(plankEndNormal)
  local length = VecLength(VecSub(plankStartTransformed,plankEnd))
  local mag = math.max(0.1,math.round(length*10)/10)
  
  local rot = GetPlankRot(plankStartTransformed, plankEnd, normal) 

  local pos = plankStartTransformed
  local tr = Transform(pos,rot)
  local pw = plankWidth/20
  local p1 = TransformToParentPoint(tr, Vec(pw, 0.0,0))
  local p2 = TransformToParentPoint(tr, Vec(-pw, 0.0,0))
  local p3 = TransformToParentPoint(tr, Vec(-pw, 0.0,-mag))
  local p4 = TransformToParentPoint(tr, Vec(pw, 0.0,-mag))

  local color = {1,1,1}
  DrawLineWithColor(p1,p2,color)
  DrawLineWithColor(p2,p3,color)
  DrawLineWithColor(p3,p4,color)
  DrawLineWithColor(p4,p1,color)
end


function DrawLineWithColor(p1,p2,color,debug)
  local r = color[1] or 1
  local g = color[2] or 1
  local b = color[3] or 1
  local a = color[4] or 1
  if debug then
    DebugLine(p1,p2,r,g,b,a)
  else
    DrawLine(p1,p2,r,g,b,a)
  end
  
end


function UpdateTransformedStart() 
  local shapetf = GetShapeWorldTransform(plankStartShape)
  plankStartTransformed = TransformToParentPoint(shapetf, plankStart)
  plankStartNormalTransformed = TransformToParentVec(shapetf, plankStartNormal)
end


function GetCorrectPlankNormal(normal)
  if VecDot(normal,plankStartNormalTransformed) < 0.75 then 
    normal = Vec(0,0,0) -- normal is zero'd out so that redundant code is not run
  end
  return normal
end


function GetPlankRot(from, to, normal) -- slightly jank, but at least it works
  local rot = QuatLookAt(from,to)
  local forward = VecNormalize(VecSub(from,to))

  local newrot = rot
  if VecLength(normal) > 0 then
    local newnormal = TransformToLocalVec(Transform(Vec(0,0,0), rot), normal)
    newnormal[3] = 0 -- zero out z and normalize
    newnormal = VecNormalize(newnormal)
    newrot = QuatRotateQuat(QuatAxisAngle(forward, math.deg(-math.atan2(newnormal[1], newnormal[2])) ),  rot)
  end

  return newrot
end


function SpawnPlank(from,to,normal)
  local length = math.max(1,math.round(VecLength(VecSub(from,to))*10))
  local rot = GetPlankRot(from, to, normal)
  
  local pw = plankWidth/20
  local joint1 = "0.0 0.05 " .. pw
  local joint2 = "" .. length/10 .. " 0.05 " .. pw
  local boxrot = "0 90 0"
  local style = plankStyles[plankIndex]
  local joint2string = ''
  if plankConnects then
    joint2string = '<joint sound="true" pos="'..joint2..'" size="0.25" collide="false" rotspring="0" rotstrength="0.015" type="ball" />'
  end
  local plankXML = '\
    <voxbox \
    tags="' .. style.tags .. '" \
    brush="'.. style.brush ..'" \
    pos="'..-pw..' -0.05 0" \
    collide="true" \
    prop="true" \
    strength="' .. style.strength .. '" \
    rot="'..boxrot..'" \
    color="'.. style.color .. '" \
    size="' .. length .. ' ' .. plankHeight .. ' ' .. plankWidth .. '" \
    texture="6 0.5" \
    material="' .. style.mat .. '" \
    pbr="0 0 0 0" \
    >\
    <joint sound="true" pos="'..joint1..'" size="0.25" collide="false" rotspring="0" rotstrength="0.015" type="ball" />\
    ' .. joint2string .. '\
    </voxbox>'
    
  Spawn(plankXML,Transform(from,rot),true,true)
end


function math.round(val)
  if val >= 0 then
    return math.floor(val + 0.5)
  else 
    return math.ceil(val - 0.5)
  end
end


function KeyPressCheck(name)
  local shift = InputDown("shift")
  local alt = InputDown("alt")
  local ctrl = InputDown("ctrl")
  local key = string.upper(InputLastPressedKey())
  local path = SETTINGS_BASE .. ".binds." .. name

  if key == string.upper(GetString(path .. ".key"))
  and shift == GetBool(path .. ".shift")
  and alt == GetBool(path .. ".alt")
  and ctrl == GetBool(path .. ".ctrl") then
    return true
  else
    return false
  end
end

function KeyDownCheck(name)
  local path = SETTINGS_BASE .. ".binds." .. name
  local key = GetString(path .. ".key")
  local shift = GetBool(path .. ".shift")
  local alt = GetBool(path .. ".alt")
  local ctrl = GetBool(path .. ".ctrl")

  if InputDown(key)
  and shift == InputDown("shift")
  and alt == InputDown("alt")
  and ctrl == InputDown("ctrl") then
    return true
  else
    return false
  end
end


function KeyPressed(name)
  -- print("key pressed:", name)
  if not name then
    keyPressedLastWaiting = true
  end
  
  if KeyPressCheck("width_sub") or name == "width_sub" then
    plankWidth = plankWidth - 1
    name = "width_sub"
  end
  if KeyPressCheck("width_add") or name == "width_add"  then
    plankWidth = plankWidth + 1
    name = "width_add"
  end
  if KeyPressCheck("height_sub") or name == "height_sub"  then
    plankHeight = plankHeight - 1
    name = "height_sub"
  end
  if KeyPressCheck("height_add") or name == "height_add" then
    plankHeight = plankHeight + 1
    name = "height_add"
  end
  if plankWidth < plankWidthMin then
    plankWidth = plankWidthMin
    PlaySound(soundError)
  end
  if plankWidth > plankWidthMax then
    plankWidth = plankWidthMax
    PlaySound(soundError)
  end
  if plankHeight < plankHeightMin then
    plankHeight = plankHeightMin
    PlaySound(soundError)
  end
  if plankHeight > plankHeightMax then
    plankHeight = plankHeightMax
    PlaySound(soundError)
  end
  SetInt(SETTINGS_BASE .. ".width", plankWidth)
  SetInt(SETTINGS_BASE .. ".height", plankHeight)
  SetString("hud.notification",plankWidth .. " x " .. plankHeight)
  
  if name then
    keyPressedLast = name
    keyPressedLastTime = 0
    -- print(keyPressedLast)
  end
  
end


function draw()
  if GetString("game.player.tool") == "infiniteplank" and showPopup then 
    local w = 320
    local h = 200
    local p = 20
    local bh = 14
    UiTranslate(UiWidth()-w-p,p)
    UiAlign("left top")
    UiColor(0,0,0,0.8)
    UiRect(w,h)
    UiPush()
      UiTranslate(w/2,40)
      UiColor(1,1,1,0.9)
      UiAlign("center middle")
      UiFont("bold.ttf",30)
      UiText("Replace original Plank?")

      UiFont("bold.ttf",20)
      UiColor(1,1,1,0.4)
      UiTranslate(0,35)
      UiText("One time question")
      UiTranslate(0,25)
      UiText("You can change it in options later")

    UiPop()
    
    UiPush()
      UiAlign("left bottom")
      UiTranslate(p,h-p)
      UiColor(0,1,0,0.9)
      UiFont("bold.ttf",25)
      UiButtonImageBox("ui/common/box-solid-6.png", 14, 14, 0, 1, 0, 0.2)
      UiTextButton("[ Y ] E S")
    UiPop()
    
    UiPush()
      UiAlign("right bottom")
      UiTranslate(w-p,h-p)
      UiColor(1,0,0,0.9)
      UiFont("bold.ttf",25)
      UiButtonImageBox("ui/common/box-solid-6.png", 14, 14, 1, 0, 0, 0.2)
      UiTextButton("[ N ] O")
    UiPop()
    
  end
end


